
/***************************************************************************
 * common_modified.cc --  This file holds all those functions and classes  *
 * that have been reused from nmap's code but that needed to be modified   *
 * in order to reuse them.                                                 *
 *                                                                         *
 ***********************IMPORTANT NMAP LICENSE TERMS************************
 *                                                                         *
 * The Nmap Security Scanner is (C) 1996-2009 Insecure.Com LLC. Nmap is    *
 * also a registered trademark of Insecure.Com LLC.  This program is free  *
 * software; you may redistribute and/or modify it under the terms of the  *
 * GNU General Public License as published by the Free Software            *
 * Foundation; Version 2 with the clarifications and exceptions described  *
 * below.  This guarantees your right to use, modify, and redistribute     *
 * this software under certain conditions.  If you wish to embed Nmap      *
 * technology into proprietary software, we sell alternative licenses      *
 * (contact sales@insecure.com).  Dozens of software vendors already       *
 * license Nmap technology such as host discovery, port scanning, OS       *
 * detection, and version detection.                                       *
 *                                                                         *
 * Note that the GPL places important restrictions on "derived works", yet *
 * it does not provide a detailed definition of that term.  To avoid       *
 * misunderstandings, we consider an application to constitute a           *
 * "derivative work" for the purpose of this license if it does any of the *
 * following:                                                              *
 * o Integrates source code from Nmap                                      *
 * o Reads or includes Nmap copyrighted data files, such as                *
 *   nmap-os-db or nmap-service-probes.                                    *
 * o Executes Nmap and parses the results (as opposed to typical shell or  *
 *   execution-menu apps, which simply display raw Nmap output and so are  *
 *   not derivative works.)                                                *
 * o Integrates/includes/aggregates Nmap into a proprietary executable     *
 *   installer, such as those produced by InstallShield.                   *
 * o Links to a library or executes a program that does any of the above   *
 *                                                                         *
 * The term "Nmap" should be taken to also include any portions or derived *
 * works of Nmap.  This list is not exclusive, but is meant to clarify our *
 * interpretation of derived works with some common examples.  Our         *
 * interpretation applies only to Nmap--we don't speak for other people's  *
 * GPL works.                                                              *
 *                                                                         *
 * If you have any questions about the GPL licensing restrictions on using *
 * Nmap in non-GPL works, we would be happy to help.  As mentioned above,  *
 * we also offer alternative license to integrate Nmap into proprietary    *
 * applications and appliances.  These contracts have been sold to dozens  *
 * of software vendors, and generally include a perpetual license as well  *
 * as providing for priority support and updates as well as helping to     *
 * fund the continued development of Nmap technology.  Please email        *
 * sales@insecure.com for further information.                             *
 *                                                                         *
 * As a special exception to the GPL terms, Insecure.Com LLC grants        *
 * permission to link the code of this program with any version of the     *
 * OpenSSL library which is distributed under a license identical to that  *
 * listed in the included COPYING.OpenSSL file, and distribute linked      *
 * combinations including the two. You must obey the GNU GPL in all        *
 * respects for all of the code used other than OpenSSL.  If you modify    *
 * this file, you may extend this exception to your version of the file,   *
 * but you are not obligated to do so.                                     *
 *                                                                         *
 * If you received these files with a written license agreement or         *
 * contract stating terms other than the terms above, then that            *
 * alternative license agreement takes precedence over these comments.     *
 *                                                                         *
 * Source is provided to this software because we believe users have a     *
 * right to know exactly what a program is going to do before they run it. *
 * This also allows you to audit the software for security holes (none     *
 * have been found so far).                                                *
 *                                                                         *
 * Source code also allows you to port Nmap to new platforms, fix bugs,    *
 * and add new features.  You are highly encouraged to send your changes   *
 * to nmap-dev@insecure.org for possible incorporation into the main       *
 * distribution.  By sending these changes to Fyodor or one of the         *
 * Insecure.Org development mailing lists, it is assumed that you are      *
 * offering the Nmap Project (Insecure.Com LLC) the unlimited,             *
 * non-exclusive right to reuse, modify, and relicense the code.  Nmap     *
 * will always be available Open Source, but this is important because the *
 * inability to relicense code has caused devastating problems for other   *
 * Free Software projects (such as KDE and NASM).  We also occasionally    *
 * relicense the code to third parties as discussed above.  If you wish to *
 * specify special license conditions of your contributions, just say so   *
 * when you send them.                                                     *
 *                                                                         *
 * This program is distributed in the hope that it will be useful, but     *
 * WITHOUT ANY WARRANTY; without even the implied warranty of              *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       *
 * General Public License v2.0 for more details at                         *
 * http://www.gnu.org/licenses/gpl-2.0.html , or in the COPYING file       *
 * included with Nmap.                                                     *
 *                                                                         *
 ***************************************************************************
 * The code in this file was originally written by Luis MartinGarcia       *
 * during Google Summer of Code programme, 2009.                           *
 *                                                                         *
 ***************************************************************************/
#include "nping.h"
#include "output.h"
#include "dnet.h"
#include "pcap.h"
#include "common.h"
#include "utils.h"
#include "utils_net.h"
#include "common_modified.h"
#include "NpingOps.h"
#ifdef WIN32
#include "pcap-int.h"
#endif


extern NpingOps o;

/*****************************************************************************
  * STUFF FROM TargetGroup.cc
  ****************************************************************************

  CHANGES:
       Modified parse_expr.
       Modified get_next_host:

*/


TargetGroup::TargetGroup() {
  Initialize();
}

// Bring back (or start with) original state
void TargetGroup::Initialize() {
  targets_type = TYPE_NONE;
  memset(addresses, 0, sizeof(addresses));
  memset(current, 0, sizeof(current));
  memset(last, 0, sizeof(last));
  ipsleft = 0;
}

/* take the object back to the beginning without  (mdmcl)
 * reinitalizing the data structures */
int TargetGroup::rewind() {

  /* For netmasks we must set the current address to the
   * starting address and calculate the ips by distance */
  if (targets_type == IPV4_NETMASK) {
    currentaddr = startaddr;
    if (startaddr.s_addr <= endaddr.s_addr) {
      ipsleft = ((unsigned long long) (endaddr.s_addr - startaddr.s_addr)) + 1;
      return 0;
    }
    else
      assert(0);
  }
  /* For ranges, we easily set current to zero and calculate
   * the ips by the number of values in the columns */
  else if (targets_type == IPV4_RANGES) {
    memset((char *)current, 0, sizeof(current));
    ipsleft = (unsigned long long) (last[0] + 1) *
              (unsigned long long) (last[1] + 1) *
              (unsigned long long) (last[2] + 1) *
              (unsigned long long) (last[3] + 1);
    return 0;
  }
#if HAVE_IPV6
  /* For IPV6 there is only one address, this function doesn't
   * make much sence for IPv6 does it? */
  else if (targets_type == IPV6_ADDRESS) {
    ipsleft = 1;
    return 0;
  }
#endif

  /* If we got this far there must be an error, wrong type */
  return -1;
}



/* For ranges, skip all hosts in an octet,                  (mdmcl)
 * get_next_host should be used for skipping the last octet :-)
 * returns: number of hosts skipped */
int TargetGroup::skip_range(_octet_nums octet) {
  unsigned long hosts_skipped = 0, /* number of hosts skipped */
      oct = 0;           /* octect number */
      int i = 0;                 /* simple lcv */

  /* This function is only supported for RANGES! */
  if (targets_type != IPV4_RANGES)
    return -1;

  switch (octet) {
    case FIRST_OCTET:
      oct = 0;
      hosts_skipped = (last[1] + 1) * (last[2] + 1) * (last[3] + 1);
      break;
    case SECOND_OCTET:
      oct = 1;
      hosts_skipped = (last[2] + 1) * (last[3] + 1);
      break;
    case THIRD_OCTET:
      oct = 2;
      hosts_skipped = (last[3] + 1);
      break;
    default:  /* Hmm, how'd you do that */
      return -1;
  }

  /* catch if we try to take more than are left */
  assert(ipsleft + 1>= hosts_skipped);

  /* increment the next octect that we can above us */
  for (i = oct; i >= 0; i--) {
    if (current[i] < last[i]) {
      current[i]++;
      break;
    }
    else
      current[i] = 0;
  }

  /* reset all the ones below us to zero */
  for (i = oct+1; i <= 3; i++) {
    current[i] = 0;
  }

  /* we actually don't skip the current, it was accounted for
   * by get_next_host */
  ipsleft -= hosts_skipped - 1;

  return hosts_skipped;
}

 /* Grab the next host from this expression (if any) and updates its internal
    state to reflect that the IP was given out.  Returns 0 and
    fills in ss if successful.  ss must point to a pre-allocated
    sockaddr_storage structure */
int TargetGroup::get_next_host(struct sockaddr_storage *ss, size_t *sslen) {

  int octet;
  struct sockaddr_in *sin = (struct sockaddr_in *) ss;
  struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *) ss;

  /* CHANGE: Commented out. See note at the end of the method */
  //startover: /* to handle nmap --resume where I have already
  //            * scanned many of the IPs */
  assert(ss);
  assert(sslen);


  if (ipsleft == 0)
    return -1;

  if (targets_type == IPV4_NETMASK) {
    memset(sin, 0, sizeof(struct sockaddr_in));
    sin->sin_family = AF_INET;
    *sslen = sizeof(struct sockaddr_in);
#if HAVE_SOCKADDR_SA_LEN
    sin->sin_len = *sslen;
#endif

    if (currentaddr.s_addr <= endaddr.s_addr) {
      sin->sin_addr.s_addr = htonl(currentaddr.s_addr++);
    } else {
      error("Bogus target structure passed to %s", __func__);
      ipsleft = 0;
      return -1;
    }
  }
  else if (targets_type == IPV4_RANGES) {
    memset(sin, 0, sizeof(struct sockaddr_in));
    sin->sin_family = AF_INET;
    *sslen = sizeof(struct sockaddr_in);
#if HAVE_SOCKADDR_SA_LEN
    sin->sin_len = *sslen;
#endif
    //if (o.debugging > 2) { /* CHANGE: Do not use NmapOps and do not use log_Write*/
    //  log_write(LOG_STDOUT, "doing %d.%d.%d.%d = %d.%d.%d.%d\n", current[0], current[1], current[2], current[3], addresses[0][current[0]],addresses[1][current[1]],addresses[2][current[2]],addresses[3][current[3]]);
    //}
    //outPrint(DBG_2, "doing %d.%d.%d.%d = %d.%d.%d.%d", current[0], current[1], current[2], current[3], addresses[0][current[0]],addresses[1][current[1]],addresses[2][current[2]],addresses[3][current[3]]);


    /* Set the IP to the current value of everything */
    sin->sin_addr.s_addr = htonl(addresses[0][current[0]] << 24 |
                                 addresses[1][current[1]] << 16 |
                                 addresses[2][current[2]] <<  8 |
                                 addresses[3][current[3]]);

    /* Now we nudge up to the next IP */
    for(octet = 3; octet >= 0; octet--) {
      if (current[octet] < last[octet]) {
        /* OK, this is the column I have room to nudge upwards */
        current[octet]++;
        break;
      } else {
        /* This octet is finished so I reset it to the beginning */
        current[octet] = 0;
      }
    }
    if (octet == -1) {
      /* It didn't find anything to bump up, I must have taken the last IP */
      assert(ipsleft == 1);
      /* So I set current to last with the very final octet up one ... */
      /* Note that this may make current[3] == 256 */
      current[0] = last[0]; current[1] = last[1];
      current[2] = last[2]; current[3] = last[3] + 1;
    } else {
      assert(ipsleft > 1); /* There must be at least one more IP left */
    }
  } else {
    assert(targets_type == IPV6_ADDRESS);
    assert(ipsleft == 1);
#if HAVE_IPV6
    *sslen = sizeof(struct sockaddr_in6);
    memset(sin6, 0, *sslen);
    sin6->sin6_family = AF_INET6;
#ifdef SIN_LEN
    sin6->sin6_len = *sslen;
#endif /* SIN_LEN */
    memcpy(sin6->sin6_addr.s6_addr, ip6.sin6_addr.s6_addr, 16);
    sin6->sin6_scope_id = ip6.sin6_scope_id;
#else
    fatal("IPV6 not supported on this platform");
#endif // HAVE_IPV6
  }
  ipsleft--;


  /* CHANGE: These lines have been commented out to make this code
   * independent from NmapOps  */
  /* If we are resuming from a previous scan, we have already finished
     scans up to o.resume_ip.  */
 // if (sin->sin_family == AF_INET && o.resume_ip.s_addr) {
 //   if (o.resume_ip.s_addr == sin->sin_addr.s_addr)
 //     o.resume_ip.s_addr = 0; /* So that we will KEEP the next one */
 //   goto startover; /* Try again */
 // }

  return 0;
}

/* Returns the last given host, so that it will be given again next
     time get_next_host is called.  Obviously, you should only call
     this if you have fetched at least 1 host since parse_expr() was
     called */
int TargetGroup::return_last_host() {
  int octet;

  ipsleft++;
  if (targets_type == IPV4_NETMASK) {
    assert(currentaddr.s_addr > startaddr.s_addr);
    currentaddr.s_addr--;
  } else if (targets_type == IPV4_RANGES) {
    for(octet = 3; octet >= 0; octet--) {
      if (current[octet] > 0) {
        /* OK, this is the column I have room to nudge downwards */
        current[octet]--;
        break;
      } else {
        /* This octet is already at the beginning, so I set it to the end */
        current[octet] = last[octet];
      }
    }
    assert(octet != -1);
  } else {
    assert(targets_type == IPV6_ADDRESS);
    assert(ipsleft == 1);
  }
  return 0;
}





/* TODO: WARNING: This functions has been modified for portability. Check
 * for label "CHANGE:" in the code to see the actual changes.
 *
 * UPDATE: Added support for DNS resolution caching. Using function
 * gethostbynameCached() instead of gethostbyname()
 */
 /* Initializes (or reinitializes) the object with a new expression, such
    as 192.168.0.0/16 , 10.1.0-5.1-254 , or fe80::202:e3ff:fe14:1102 .
    Returns 0 for success */
int TargetGroup::parse_expr(const char * const target_expr, int af) {

  int i=0,j=0,k=0;
  int start, end;
  char *r,*s, *target_net;
  char *addy[5];
  char *hostexp = strdup(target_expr);
  struct hostent *target;
  namedhost = 0;

  if (targets_type != TYPE_NONE)
    Initialize();

  ipsleft = 0;

  if (af == AF_INET) {

    if (strchr(hostexp, ':'))
      fatal("Invalid host expression: %s -- colons only allowed in IPv6 addresses, and then you need the -6 switch", hostexp);

    /*struct in_addr current_in;*/
    addy[0] = addy[1] = addy[2] = addy[3] = addy[4] = NULL;
    addy[0] = r = hostexp;
    /* First we break the expression up into the four parts of the IP address
       + the optional '/mask' */
    target_net = hostexp;
    s = strchr(hostexp, '/'); /* Find the slash if there is one */
    if (s) {
      char *tail;
      long netmask_long;

      *s = '\0';  /* Make sure target_net is terminated before the /## */
      s++;        /* Point s at the netmask */
      if (!isdigit(*s)) {
        error("Illegal netmask value, must be /0 - /32 .  Assuming /32 (one host)");
        netmask = 32;
      } else {
        netmask_long = strtol(s, (char**) &tail, 10);
        if (*tail != '\0' || tail == s || netmask_long < 0 || netmask_long > 32) {
          error("Illegal netmask value, must be /0 - /32 .  Assuming /32 (one host)");
          netmask = 32;
        } else
          netmask = (u32) netmask_long;
      }
    } else
      netmask = 32;
    for(i=0; *(hostexp + i); i++)
      if (isupper((int) *(hostexp +i)) || islower((int) *(hostexp +i))) {
        namedhost = 1;
        break;
      }
    if (netmask != 32 || namedhost) {
      targets_type = IPV4_NETMASK;
      if (!inet_pton(AF_INET, target_net, &(startaddr))) {

        /* There is a bug report on the use of gethostbynameCached()
         * <http://seclists.org/nmap-dev/2010/q1/803>
         * I haven't been able to find any problem with that code but
         * still, the fact that DNS queries are cached does not improve
         * performance a lot. It may save one DNS query per execution
         * in those cases wher NpingOps::validateOptions() grabs the
         * first target and uses it to determine output network interface.
         * It would also save some queries in the case where a user
         * specified the same host twice in the commandlined, something
         * that does not make much sense anyway. However, since the call
         * to gethostbynameCached() seems to cause denial of service
         * for some people, I think it's ok to disable its use for now
         * and enable it later if there is a good reason for it.
         *
         * Luis MartinGarcia. */
        //if ((target = gethostbynameCached(target_net))) {
        if ((target = gethostbyname(target_net))) {
          int count=0;

          memcpy(&(startaddr), target->h_addr_list[0], sizeof(struct in_addr));

          while (target->h_addr_list[count]) count++;

          if (count > 1)
             outPrint(DBG_2,"Warning: Hostname %s resolves to %d IPs. Using %s.", target_net, count, inet_ntoa(*((struct in_addr *)target->h_addr_list[0])));
        } else {
          error("Failed to resolve given hostname/IP: %s.  Note that you can't use '/mask' AND '1-4,7,100-' style IP ranges", target_net);
          free(hostexp);
          return 1;
        }
      }
      if (netmask) {
        unsigned long longtmp = ntohl(startaddr.s_addr);
        startaddr.s_addr = longtmp & (unsigned long) (0 - (1<<(32 - netmask)));
        endaddr.s_addr = longtmp | (unsigned long)  ((1<<(32 - netmask)) - 1);
      } else {
        /* The above calculations don't work for a /0 netmask, though at first
         * glance it appears that they would
         */
        startaddr.s_addr = 0;
        endaddr.s_addr = 0xffffffff;
      }
      currentaddr = startaddr;
      if (startaddr.s_addr <= endaddr.s_addr) {
        ipsleft = ((unsigned long long) (endaddr.s_addr - startaddr.s_addr)) + 1;
        free(hostexp);
        return 0;
      }
      fprintf(stderr, "Host specification invalid");
      free(hostexp);
      return 1;
    }
    else {
      targets_type = IPV4_RANGES;
      i=0;

      while(*r) {
        if (*r == '.' && ++i < 4) {
          *r = '\0';
          addy[i] = r + 1;
        }
        else if (*r != '*' && *r != ',' && *r != '-' && !isdigit((int)*r))
          fatal("Invalid character in host specification.  Note in particular that square brackets [] are no longer allowed.  They were redundant and can simply be removed.");
        *r++;
      }
      if (i != 3) fatal("Invalid target host specification: %s", target_expr);

      for(i=0; i < 4; i++) {
        j=0;
        do {
          s = strchr(addy[i],',');
          if (s) *s = '\0';
          if (*addy[i] == '*') { start = 0; end = 255; }
          else if (*addy[i] == '-') {
            start = 0;
            if (*(addy[i] + 1) == '\0') end = 255;
            else end = atoi(addy[i]+ 1);
          }
          else {
            start = end = atoi(addy[i]);
            if ((r = strchr(addy[i],'-')) && *(r+1) ) end = atoi(r + 1);
            else if (r && !*(r+1)) end = 255;
          }
       /* if (o.debugging > 2)
        *   log_write(LOG_STDOUT, "The first host is %d, and the last one is %d\n", start, end); */
          if (start < 0 || start > end || start > 255 || end > 255)
            fatal("Your host specifications are illegal!");
          if (j + (end - start) > 255)
            fatal("Your host specifications are illegal!");
          for(k=start; k <= end; k++)
            addresses[i][j++] = k;
          last[i] = j-1;
          if (s) addy[i] = s + 1;
        } while (s);
      }
    }
    memset((char *)current, 0, sizeof(current));
    ipsleft = (unsigned long long) (last[0] + 1) *
              (unsigned long long) (last[1] + 1) *
              (unsigned long long) (last[2] + 1) *
              (unsigned long long) (last[3] + 1);
    }
  else {
#if HAVE_IPV6
    int rc = 0;
    assert(af == AF_INET6);
    if (strchr(hostexp, '/')) {
      fatal("Invalid host expression: %s -- slash not allowed.  IPv6 addresses can currently only be specified individually", hostexp);
    }
    targets_type = IPV6_ADDRESS;
    struct addrinfo hints;
    struct addrinfo *result = NULL;
    memset(&hints, 0, sizeof(hints));
    hints.ai_family = PF_INET6;
    rc = getaddrinfo(hostexp, NULL, &hints, &result);
    if (rc != 0 || result == NULL) {
      error("Failed to resolve given IPv6 hostname/IP: %s.  Note that you can't use '/mask' or '[1-4,7,100-]' style ranges for IPv6.  Error code %d: %s", hostexp, rc, gai_strerror(rc));
      free(hostexp);
      if (result) freeaddrinfo(result);
      return 1;
    }
    assert(result->ai_addrlen == sizeof(struct sockaddr_in6));
    struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *) result->ai_addr;
    memcpy(&ip6, sin6, sizeof(struct sockaddr_in6));
    ipsleft = 1;
    freeaddrinfo(result);
#else // HAVE_IPV6
    fatal("IPv6 not supported on your platform");
#endif // HAVE_IPV6
  }

  free(hostexp);
  return 0;
}










/*****************************************************************************/





/*****************************************************************************/
/* TODO: Place this in nbase and update nmap's code? */
/* Changed a bit from the original to make it independent of NmapOps
 * CHANGES: Added parameter random. If true, a string containing a random,
 *          non reserved, IP address will be returned.                       */
char *grab_next_host_spec(FILE *inputfd, bool random, int argc, char **fakeargv) {
  static char host_spec[1024];
  unsigned int host_spec_index;
  int ch;
  struct in_addr ip;

  if (random) {
    do {
      ip.s_addr = get_random_u32();
    } while (ip_is_reserved(&ip));
    Strncpy(host_spec, inet_ntoa(ip), sizeof(host_spec));
  } else if (!inputfd) {
    return( (optind < argc)?  fakeargv[optind++] : NULL);
  } else {
    host_spec_index = 0;
    while((ch = getc(inputfd)) != EOF) {
      if (ch == ' ' || ch == '\r' || ch == '\n' || ch == '\t' || ch == '\0') {
	if (host_spec_index == 0) continue;
	host_spec[host_spec_index] = '\0';
	return host_spec;
      } else if (host_spec_index < sizeof(host_spec) / sizeof(char) -1) {
	host_spec[host_spec_index++] = (char) ch;
      } else fatal("One of the host_specifications from your input file is too long (> %d chars)", (int) sizeof(host_spec));
    }
    host_spec[host_spec_index] = '\0';
  }
  if (!*host_spec) return NULL;
  return host_spec;

} /* End of grab_next_host_spec() */
/*****************************************************************************/






/*****************************************************************************/
/* Send a pre-built IPv4 packet */
/* CHANGED to make it NmapOps independent */
int send_ip_packet(int sd, struct eth_nfo *eth, u8 *packet, unsigned int packetlen) {
  //struct ip *ip = (struct ip *) packet;
  int res;

  assert(packet);
  assert((int) packetlen > 0);

  // fragmentation requested && packet is bigger than MTU
  //if (o.fragscan && (packetlen - ip->ip_hl * 4 > (unsigned int) o.fragscan))
  //  return send_frag_ip_packet(sd, eth, packet, packetlen, o.fragscan);

  if (eth)
    res = send_ip_packet_eth(eth, packet, packetlen);
  else
    res = send_ip_packet_sd(sd, packet, packetlen);

  //PacketTrace::trace(PacketTrace::SENT, packet, packetlen); 

  return res;
}








/*****************************************************************************/
/* It was static but I removed the "static" label */
/* Also: changed reference to o.debugging
 * -    if (o.debugging > 1 || numerrors <= 10) {
 * +    if (o.getDebugging() > 1 || numerrors <= 10) {
 * Commented out packet trace line.
 * Added IPv6 compatibility
 * */
int Sendto(const char *functionname, int sd, const unsigned char *packet,
		  int len, unsigned int flags, struct sockaddr *to, int tolen) {

struct sockaddr_in *sin4 = (struct sockaddr_in *)to;
struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *)to;
int res;
int retries = 0;
int sleeptime = 0;
static int numerrors = 0;
char ipstring[128];

assert(packet);
assert(to);
assert(tolen>0);
assert(len>0);

do {
  if ((res = sendto(sd, (const char *) packet, len, flags, to, tolen)) == -1) {
    int err = socket_errno();

    numerrors++;

    if( sin4->sin_family==AF_INET ){
        inet_ntop(AF_INET, &sin4->sin_addr, ipstring, sizeof(ipstring));
    }
    else if( sin6->sin6_family==AF_INET6){
        inet_ntop(AF_INET6, &sin6->sin6_addr, ipstring, sizeof(ipstring));
    }
    
    if (o.getDebugging() > 1 || numerrors <= 10) {
      error("sendto in %s: sendto(%d, packet, %d, 0, %s, %d) => %s",
	    functionname, sd, len, ipstring, tolen,
	    strerror(err));
      error("Offending packet: %s", ippackethdrinfo(packet, len, HIGH_DETAIL));
      if (numerrors == 10) {
	error("Omitting future %s error messages now that %d have been shown.  Use -d2 if you really want to see them.", __func__, numerrors);
      }
    }

#if WIN32
	return -1;
#else
    if (retries > 2 || err == EPERM || err == EACCES || err == EADDRNOTAVAIL
	|| err == EINVAL)
      return -1;
    sleeptime = 15 * (1 << (2 * retries));
    error("Sleeping %d seconds then retrying", sleeptime);
    fflush(stderr);
    sleep(sleeptime);
#endif
  }
  retries++;
} while( res == -1);

 //PacketTrace::trace(PacketTrace::SENT, packet, len);

return res;
}
/*****************************************************************************/









/*****************************************************************************/
/* getpts() and getpts_simple() (see above) are wrappers for this function */
void getpts_aux(const char *origexpr, int nested, u8 *porttbl, int *portwarning) {
  long rangestart = -2343242, rangeend = -9324423;
  const char *current_range;
  char *endptr;
  //char servmask[128];  // A protocol name can be up to 127 chars + nul byte
  //int i;

  /* An example of proper syntax to use in error messages. */
  const char *syntax_example;
  //if (change_range_type)
  //  syntax_example = "-100,200-1024,T:3000-4000,U:60000-";
  //else
    syntax_example = "-100,200-1024,3000-4000,60000-";

  current_range = origexpr;
  do {
    while(isspace((int) *current_range))
      current_range++; /* I don't know why I should allow spaces here, but I will */

    //if (change_range_type) {
      //if (*current_range == 'T' && *++current_range == ':') {
          //current_range++;
          //range_type = SCAN_TCP_PORT;
          //continue;
      //}
      //if (*current_range == 'U' && *++current_range == ':') {
          //current_range++;
          //range_type = SCAN_UDP_PORT;
          //continue;
      //}
      //if (*current_range == 'S' && *++current_range == ':') {
          //current_range++;
          //range_type = SCAN_SCTP_PORT;
          //continue;
      //}
      //if (*current_range == 'P' && *++current_range == ':') {
          //current_range++;
          //range_type = SCAN_PROTOCOLS;
          //continue;
      //}
    //}

    if (*current_range == '[') {
      if (nested)
        fatal("Can't nest [] brackets in port/protocol specification");

      //getpts_aux(++current_range, 1, porttbl, range_type, portwarning);
        getpts_aux(++current_range, 1, porttbl, portwarning); // ADDED

      // Skip past the ']'. This is OK because we can't nest []s
      while(*current_range != ']') current_range++;
      current_range++;

      // Skip over a following ',' so we're ready to keep parsing
      if (*current_range == ',') current_range++;

      continue;
    } else if (*current_range == ']') {
      if (!nested)
        fatal("Unexpected ] character in port/protocol specification");

      return;
    } else if (*current_range == '-') {
      //if (range_type & SCAN_PROTOCOLS)
      //  rangestart = 0;
      //else
        rangestart = 1;
    }
    else if (isdigit((int) *current_range)) {
      rangestart = strtol(current_range, &endptr, 10);
      //if (range_type & SCAN_PROTOCOLS) {
      //  if (rangestart < 0 || rangestart > 255)
	  //fatal("Protocols to be scanned must be between 0 and 255 inclusive");
      //} else {
        if (rangestart < 0 || rangestart > 65535)
	        fatal("Ports to be scanned must be between 0 and 65535 inclusive");
      //}
      current_range = endptr;
      while(isspace((int) *current_range)) current_range++;
    } //else if (islower((int) *current_range) || *current_range == '*' || *current_range == '?') {
      //i = 0;

      //while (*current_range && !isspace((int)*current_range) && *current_range != ',' && *current_range != ']') {
      //  servmask[i++] = *(current_range++);
      //  if (i >= ((int)sizeof(servmask)-1))
      //    fatal("A service mask in the port/protocol specification is either malformed or too long");
     // }

     // if (*current_range && *current_range != ']') current_range++; // We want the '] character to be picked up on the next pass
     // servmask[i] = '\0'; // Finish the string

      //i = addportsfromservmask(servmask, porttbl, range_type);
      //if (range_type & SCAN_PROTOCOLS) i += addprotocolsfromservmask(servmask, porttbl);

      //if (i == 0)
      //  fatal("Found no matches for the service mask '%s' and your specified protocols", servmask);

      //continue;

    /*}*/ else {
      fatal("Error #485: Your port specifications are illegal.  Example of proper form: \"%s\"", syntax_example);
    }
    /* Now I have a rangestart, time to go after rangeend */
    if (!*current_range || *current_range == ',' || *current_range == ']') {
      /* Single port specification */
      rangeend = rangestart;
    } else if (*current_range == '-') {
      current_range++;
      if (!*current_range || *current_range == ',' || *current_range == ']') {
	/* Ended with a -, meaning up until the last possible port */
        //if (range_type & SCAN_PROTOCOLS)
        //  rangeend = 255;
        //else
          rangeend = 65535;
      } else if (isdigit((int) *current_range)) {
	rangeend = strtol(current_range, &endptr, 10);
   //     if (range_type & SCAN_PROTOCOLS) {
//	  if (rangeend < 0 || rangeend > 255)
//	    fatal("Protocols to be scanned must be between 0 and 255 inclusive");
//	} else {
	  if (rangeend < 0 || rangeend > 65535)
	    fatal("Ports to be scanned must be between 0 and 65535 inclusive");
//	}
	current_range = endptr;
      } else {
	fatal("Error #486: Your port specifications are illegal.  Example of proper form: \"%s\"", syntax_example);
      }
      if (rangeend < rangestart) {
        //fatal("Your %s range %ld-%ld is backwards. Did you mean %ld-%ld?",
        //  (range_type & SCAN_PROTOCOLS) ? "protocol" : "port",
        //  rangestart, rangeend, rangeend, rangestart);
        fatal("Your port range %ld-%ld is backwards. Did you mean %ld-%ld?",
          rangestart, rangeend, rangeend, rangestart);     // ADDED


      }
    } else {
	fatal("Error #487: Your port specifications are illegal.  Example of proper form: \"%s\"", syntax_example);
    }

    /* Now I have a rangestart and a rangeend, so I can add these ports */
    while(rangestart <= rangeend) {
      if (porttbl[rangestart]) {
        if (!(*portwarning)) {
	        error("WARNING:  Duplicate port number(s) specified.  Are you alert enough to be using Nping?  Have some coffee or grab a RedBull(tm).");
            (*portwarning)++;
	    }
      } else {
        //if (nested) {
          //if ((range_type & SCAN_TCP_PORT) &&
              //nmap_getservbyport(htons(rangestart), "tcp")) {
            //porttbl[rangestart] |= SCAN_TCP_PORT;
          //}
          //if ((range_type & SCAN_UDP_PORT) &&
              //nmap_getservbyport(htons(rangestart), "udp")) {
            //porttbl[rangestart] |= SCAN_UDP_PORT;
          //}
          //if ((range_type & SCAN_SCTP_PORT) &&
              //nmap_getservbyport(htons(rangestart), "sctp")) {
            //porttbl[rangestart] |= SCAN_SCTP_PORT;
          //}
          //if ((range_type & SCAN_PROTOCOLS) &&
              //nmap_getprotbynum(htons(rangestart))) {
            //porttbl[rangestart] |= SCAN_PROTOCOLS;
          //}
        //} else {
          //porttbl[rangestart] |= range_type;
        //}

         porttbl[rangestart]=1; // ADDED for NPING
      }
      rangestart++;
    }

    /* Find the next range */
    while(isspace((int) *current_range)) current_range++;

    if (*current_range == ']') {
      if (!nested) fatal("Unexpected ] character in port specification");
      return;
    }

    if (*current_range && *current_range != ',') {
      fatal("Error #488: Your port specifications are illegal.  Example of proper form: \"%s\"", syntax_example);
    }
    if (*current_range == ',')
      current_range++;
  } while(current_range && *current_range);

}
/*****************************************************************************/











/*****************************************************************************/
/* WARNING: Taken from nmap's tcpip.cc but changed for reuse.
 *
 * @param device CAN NEVER be NULL or the call will fatal(). It must point to
 * a previoulsy allocated  buffer. That buffer may be empty (contain NULL
 * values) but the supplied pointer must be valid in order for this function
 * to work. We have this restriction here because originally the code used
 * NmapOps member "device" which was always a buffer, not a simple pointer
 * that can be NULL. Since we are trying to make this function independant
 * from NmapOps while keeping its original behaviour, we have to requiere
 * the caller to supply a valid buffer pointer, even if the buffer is empty.
 * BTW, make sure supplied buffer contains a NULL value in device[0] if you
 * don't want to specify a network interface name.
*/

/* Takes a destination address (dst) and tries to determine the
   source address and interface necessary to route to this address.
   If no route is found, false is returned and rnfo is undefined.  If
   a route is found, true is returned and rnfo is filled in with all
   of the routing details.  This function takes into account -S and -e
   options set by user (o.spoofsource, o.device) */
bool route_dst(const struct sockaddr_storage *const dst, struct route_nfo *rnfo,
               bool spoofsource, char *device, struct sockaddr_storage *src) {
  struct interface_info *ifaces=NULL;
  struct interface_info *iface = NULL;
  int numifaces = 0;
  struct sys_route *routes=NULL;
  //struct sockaddr_storage spoofss;
  //size_t spoofsslen;
  int numroutes = 0;
  int ifnum=0;
  int i=0;
  u32 mask=0;
  struct sockaddr_in *ifsin=NULL, *dstsin=NULL;
  dstsin = (struct sockaddr_in *)dst;
 
	
  /* Check for NULL pointer since the original code assumed o.device to be
   * always not NULL. */
  if( device==NULL || dst==NULL)
	fatal(" NULL parameter passed to route_dst().");

  /* IP version 4 *************************************************************/
  if (dstsin->sin_family == AF_INET){

	  if (spoofsource) {
		//o.SourceSockAddr(&spoofss, &spoofsslen);
		assert(src!=NULL);
		//throughout the rest of this function we only change rnfo->srcaddr if the source isnt spoofed
		memcpy(&rnfo->srcaddr, src, sizeof(rnfo->srcaddr));
		/* The device corresponding to this spoofed address should already have been
		   set elsewhere. */
		//assert(device[0] != '\0');
	  }

	  //first we check to see if the host is directly connected,
	  //if not, we have to check the routing table
	  //also, if the user specified a device we have to check the routing table
	  if( device[0] == '\0' ){
		ifaces = getinterfaces(&numifaces);
		/* I suppose that I'll first determine whether it is a direct connect
		   instance */
		for(ifnum=0; ifnum < numifaces; ifnum++) {
		  ifsin = (struct sockaddr_in *) &ifaces[ifnum].addr;
		  if (ifsin->sin_family != AF_INET) continue;
		  if (dstsin->sin_addr.s_addr == ifsin->sin_addr.s_addr &&
		  ifaces[ifnum].device_type != devt_loopback) {
		/* Trying to scan one of the machine's own interfaces -- we need to use
		   the localhost device for this */
		for(i=0; i < numifaces; i++)
		  if (ifaces[i].device_type == devt_loopback)
			break;
		if (i < numifaces) {
		  rnfo->direct_connect = true;
		  memcpy(&rnfo->ii, &ifaces[i], sizeof(rnfo->ii));
		  /* But the source address we want to use is the target addy */
		  if (!spoofsource)
			memcpy(&rnfo->srcaddr, &ifaces[ifnum].addr, sizeof(rnfo->srcaddr));
		  return true;
		}
		/* Hmmm ... no localhost -- I guess I'll just try using the device
		   itself */
		  }
		  mask = htonl((unsigned long) (0-1) << (32 - ifaces[ifnum].netmask_bits));
		  if ((ifsin->sin_addr.s_addr & mask) == (dstsin->sin_addr.s_addr & mask)) {
		rnfo->direct_connect = true;
		memcpy(&rnfo->ii, &ifaces[ifnum], sizeof(rnfo->ii));
		if (!spoofsource) {
		  memcpy(&rnfo->srcaddr, &ifaces[ifnum].addr, sizeof(rnfo->srcaddr));
		}
		return true;
		  }
		}
	  }

	  //assume the device isnt directly connected until we know it is
	  //the only case it could be is if the user specified an interface
	  rnfo->direct_connect = false;

	  if (device[0] != '\0') {
		iface = getInterfaceByName(device);
		if (!iface) {
		  fatal("Could not find interface %s which was specified by -e", device);
		}
		/* Is it directly connected? */
		mask = htonl((unsigned long) (0-1) << (32 - iface->netmask_bits));
		ifsin = (struct sockaddr_in *) &(iface->addr);
		if ((ifsin->sin_addr.s_addr & mask) == (dstsin->sin_addr.s_addr & mask)){
		  rnfo->direct_connect = true;
		  memcpy(&rnfo->ii, iface, sizeof(rnfo->ii));
		  if (!spoofsource)
			memcpy(&rnfo->srcaddr, &(iface->addr), sizeof(rnfo->srcaddr));
		  return true;
		}
	  }

	  routes = getsysroutes(&numroutes);
	  /* Now we simply go through the list and take the first match */
	  for(i=0; i < numroutes; i++) {
		if ((routes[i].dest & routes[i].netmask) == (dstsin->sin_addr.s_addr & routes[i].netmask)) {
		  /* Yay, found a matching route. */
		  if(device[0] != '\0'){
		if(strcmp(routes[i].device->devname, iface->devname) != 0) {
		  //ignore routes that arent on the device we specified!
		  continue;
		} else {
		  if (!spoofsource)
			memcpy(&rnfo->srcaddr, &(iface->addr), sizeof(rnfo->srcaddr));
		  memcpy(&rnfo->ii, iface, sizeof(rnfo->ii));
		}
		  } else {
		memcpy(&rnfo->ii, routes[i].device, sizeof(rnfo->ii));
		if (!spoofsource) {
		  memcpy(&rnfo->srcaddr, &routes[i].device->addr, sizeof(rnfo->srcaddr));
		}
		  }
		  ifsin = (struct sockaddr_in *) &rnfo->nexthop;
		  ifsin->sin_family = AF_INET;
		  ifsin->sin_addr.s_addr = routes[i].gw.s_addr;
		  return true;
		}
	  }
	  return false;
  }else{
	fatal("route_dst(): Unknown family (%d)", dstsin->sin_family);  
  }
  
  return false;
}
/*****************************************************************************/








/* For systems without SCTP in netinet/in.h, such as MacOS X */
#ifndef IPPROTO_SCTP
#define IPPROTO_SCTP 132
#endif



/*****************************************************************************/
/* Returns a buffer of ASCII information about a packet that may look
 * like "TCP 127.0.0.1:50923 > 127.0.0.1:3 S ttl=61 id=39516 iplen=40
 * seq=625950769" or "ICMP PING (0/1) ttl=61 id=39516 iplen=40".
 * Since this is a static buffer, don't use threads or call twice
 * within (say) printf().  And certainly don't try to free() it!  The
 * returned buffer is NUL-terminated
 *
 * Changed by Luis:
 *    Changed it to support different levels of detail. "detail" must be:
 *    LOW_DETAIL    (0x01): Traditional output.
 *    MEDIUM_DETAIL (0x02): More verbose than traditional.
 *    HIGH_DETAIL   (0x03): Contents of every field of the protocol headers .
 *
 */
const char *ippackethdrinfo(const u8 *packet, u32 len, int detail) {
  struct ip *ip = (struct ip *) packet; /* IPv4 header structure.            */
  struct tcp_hdr *tcp = NULL;           /* TCP header structure.             */
  struct udp_hdr *udp = NULL;           /* UDP header structure.             */
  struct sctp_hdr *sctp = NULL;         /* SCTP header structure.            */
  static char protoinfo[1024] = "";     /* Stores final info string.         */
  char ipinfo[512] = "";                /* Temp info about IP.               */
  char icmpinfo[512] = "";              /* Temp info about ICMP.             */
  char icmptype[128]="";                /* Temp info about ICMP type & code  */
  char icmpfields[256]="";              /* Temp info for various ICMP fields */
  char fragnfo[64] = "";                /* Temp info about fragmentation.    */
  char srchost[INET6_ADDRSTRLEN] = "";  /* Src IP in dot-decimal notation.   */
  char dsthost[INET6_ADDRSTRLEN] = "";  /* Dst IP in dot-decimal notation.   */
  struct in_addr saddr, daddr;          /* Src and Dst IPs in binary.        */
  char *p = NULL;                       /* Aux pointer.                      */
  int frag_off = 0;                     /* To compute IP fragment offset.    */
  int more_fragments = 0;               /* True if IP MF flag is set.        */
  int dont_fragment = 0;                /* Ture if IP DF flag is set.        */
  int reserved_flag = 0;                /* True if IP Reserved flag is set.  */
  size_t iphdrlen=0;                    /* Length of the IP (4 or 6) header  */
  u8 nextproto=0;                       /* Protocol after IP (4 or 6) header */
  
  /* Ensure IP version makes sense */
  if (ip->ip_v != 4 && ip->ip_v != 6 )
    return "BOGUS!  IP Version in packet is not 4";


  /* Ensure we end up with a valid detail number */
  if( detail!=LOW_DETAIL && detail!=MEDIUM_DETAIL && detail!=HIGH_DETAIL)
    detail=LOW_DETAIL;

/* IP INFORMATION ************************************************************/
if( ip->ip_v == 4 ){ /* IPv4 */

  if (len < 20)
    return "BOGUS!  Packet too short.";
  else{
     iphdrlen=ip->ip_hl * 4;
     nextproto=ip->ip_p;
  }

  /* Obtain IP source and destination info */
  saddr.s_addr = ip->ip_src.s_addr;
  daddr.s_addr = ip->ip_dst.s_addr;
  inet_ntop(AF_INET, &saddr, srchost, sizeof(srchost));
  inet_ntop(AF_INET, &daddr, dsthost, sizeof(dsthost));

  /* Compute fragment offset and check if flags are set */
  frag_off = 8 * (ntohs(ip->ip_off) & 8191) /* 2^13 - 1 */;
  more_fragments = ntohs(ip->ip_off) & IP_MF;
  dont_fragment = ntohs(ip->ip_off) & IP_DF;
  reserved_flag = ntohs(ip->ip_off) & IP_RF;

  /* Is this a fragmented packet? is it the last fragment? */
  if (frag_off || more_fragments) {
    Snprintf(fragnfo, sizeof(fragnfo), " frag offset=%d%s", frag_off, more_fragments ? "+" : "");
  }

  /* Create a string with information relevant to the specified level of detail */
  if( detail == LOW_DETAIL ){
      Snprintf(ipinfo, sizeof(ipinfo), "ttl=%d id=%d iplen=%d%s %s%s%s",
          ip->ip_ttl, ntohs(ip->ip_id), ntohs(ip->ip_len), fragnfo,
          ip->ip_hl==5?"":"ipopts={",
          ip->ip_hl==5?"":print_ip_options((u8*)ip + sizeof(struct ip), MIN((unsigned)(ip->ip_hl-5)*4,len-sizeof(struct ip))),
          ip->ip_hl==5?"":"}");
  }else if( detail == MEDIUM_DETAIL ){
      Snprintf(ipinfo, sizeof(ipinfo), "ttl=%d id=%d proto=%d csum=0x%04x iplen=%d%s %s%s%s",
          ip->ip_ttl, ntohs(ip->ip_id),
          ip->ip_p, ntohs(ip->ip_sum),
          ntohs(ip->ip_len), fragnfo,
          ip->ip_hl==5?"":"ipopts={",
          ip->ip_hl==5?"":print_ip_options((u8*)ip + sizeof(struct ip), MIN((unsigned)(ip->ip_hl-5)*4,len-sizeof(struct ip))),
          ip->ip_hl==5?"":"}");
  }else if( detail==HIGH_DETAIL ){
      Snprintf(ipinfo, sizeof(ipinfo), "ver=%d ihl=%d tos=0x%02x iplen=%d id=%d%s%s%s%s foff=%d%s ttl=%d proto=%d csum=0x%04x%s%s%s",
          ip->ip_v, ip->ip_hl,
          ip->ip_tos, ntohs(ip->ip_len),
          ntohs(ip->ip_id),
          (reserved_flag||dont_fragment||more_fragments) ? " flg=" : "",
          (reserved_flag)? "x" : "",
          (dont_fragment)? "D" : "",
          (more_fragments)? "M": "",
          frag_off, (more_fragments) ? "+" : "",
          ip->ip_ttl, ip->ip_p,
          ntohs(ip->ip_sum),
          ip->ip_hl==5?"":" ipopts={",
          ip->ip_hl==5?"":print_ip_options((u8*)ip + sizeof(struct ip), MIN((unsigned)(ip->ip_hl-5)*4,len-sizeof(struct ip))),
          ip->ip_hl==5?"":"}");
  }

}else{ /* IPv6 */

  /* I'd rather use a regular u8 pointer to access the IPv6 header because
   * it's surprinsingly easy to fuck the whole thing up using structures due
   * to the weird IPv6 field alignment and the f*** compiler padding structs
   * when it shouldn't. */
  u8  *ipv6pnt = (u8 *)packet;

  if (len < 40)
    return "BOGUS!  IPv6 Packet too short.";
  else{
     iphdrlen=40;
     nextproto=ipv6pnt[6];
  }

  /* Obtain IP source and destination info */
  struct in6_addr ip6_src;
  struct in6_addr ip6_dst;
  memcpy(ip6_src.s6_addr, &ipv6pnt[8], 16);
  memcpy(ip6_dst.s6_addr, &ipv6pnt[24], 16);
  inet_ntop(AF_INET6, &ip6_src, srchost, sizeof(srchost));
  inet_ntop(AF_INET6, &ip6_dst, dsthost, sizeof(dsthost));

  /* Obtain payload length, next protocol and hop limit */
  u16 *ipv6_pl = (u16 *)(&ipv6pnt[4]);
  u8  *ipv6_nh = (u8 *)(& ipv6pnt[6]);
  u8  *ipv6_hl = (u8 *)(& ipv6pnt[7]);


  /* Obtain flow label and traffic class */
  u32 *word = (u32 *)(&ipv6pnt[0]);
  u32 flow= ntohl( *word );

  u32 ip6_fl = flow & 0x000fffff;
  u32 ip6_tc = (flow & 0x0ff00000) >> 20;


  /* Create a string with information relevant to the specified level of detail */
  if( detail == LOW_DETAIL ){
      Snprintf(ipinfo, sizeof(ipinfo), "hopl=%d flow=%x payloadlen=%d",
          (*ipv6_hl), ip6_fl, ntohs(*ipv6_pl)
          );
  }else if( detail == MEDIUM_DETAIL ){
      Snprintf(ipinfo, sizeof(ipinfo), "hopl=%d tclass=%d flow=%x payloadlen=%d",
          (*ipv6_hl), ip6_tc, ip6_fl, ntohs(*ipv6_pl)
          );
  }else if( detail==HIGH_DETAIL ){
      Snprintf(ipinfo, sizeof(ipinfo), "ver=6, tclass=%x flow=%x payloadlen=%d nh=%s hopl=%d ",
          ip6_tc, ip6_fl, ntohs(*ipv6_pl),
          nexthdrtoa(*ipv6_nh, true), *ipv6_hl
          );
  }

}


/* TCP INFORMATION ***********************************************************/
  if (nextproto == IPPROTO_TCP) {
    char tflags[10];
    char tcpinfo[64] = "";
    char buf[32];
    char tcpoptinfo[256] = "";
    tcp = (struct tcp_hdr *)  (packet + iphdrlen);


    /* Let's parse the TCP header. The following code is very ugly because we
     * have to deal with a lot of different situations. We don't want to
     * segfault so we have to check every length and every bound to ensure we
     * don't read past the packet. We cannot even trust the contents of the
     * received packet because, for example, an IPv4 header may state it
     * carries a TCP packet but may actually carry nothing at all.
     *
     * So we distinguish 4 situations. I know the first two are weird but they
     * were there when I modified this code so I left them there just in
     * case.
     *      1. IP datagram is very small or is a fragment where we are missing
     *         the first part of the TCP header
     *      2. IP datagram is a fragment and although we are missing the first
     *         8 bytes of the TCP header, we have the rest of it (or some of
     *         the rest of it)
     *      3. IP datagram is NOT a fragment but we don't have the full TCP
     *         header, we are missing some bytes.
     *      4. IP datagram is NOT a fragment and we have at least a full 20
     *         byte TCP header.
     */


    /* CASE 1: where we don't have the first 8 bytes of the TCP header because
     * either the fragment belongs to somewhere past that or the IP contains
     * less than 8 bytes. This also includes empty IP packets that say they
     * contain a TCP packet. */
    if ( (frag_off > 8) || (len < (u32) iphdrlen + 8) ){    
      Snprintf(protoinfo, sizeof(protoinfo), "TCP %s:?? > %s:?? ?? %s (incomplete)",
        srchost, dsthost, ipinfo);
    }

    /* CASE 2: where we are missing the first 8 bytes of the TCP header but we
     * have, at least, the next 8 bytes so we can see the ACK number, the
     * flags and window size. */      
    else if ( (frag_off == 8) && (len >= (u32) iphdrlen + 8)) {

      tcp = (struct tcp_hdr *)((u8 *) tcp - frag_off); // ugly?

      /* TCP Flags */
      p = tflags;
      /* These are basically in tcpdump order */
      if (tcp->th_flags & TH_SYN) *p++ = 'S';
      if (tcp->th_flags & TH_FIN) *p++ = 'F';
      if (tcp->th_flags & TH_RST) *p++ = 'R';
      if (tcp->th_flags & TH_PUSH) *p++ = 'P';
      if (tcp->th_flags & TH_ACK){ *p++ = 'A';
            Snprintf(tcpinfo, sizeof(tcpinfo), " ack=%lu",
                (unsigned long) ntohl(tcp->th_ack));
      }
      if (tcp->th_flags & TH_URG) *p++ = 'U';
      if (tcp->th_flags & TH_ECE) *p++ = 'E'; /* rfc 2481/3168 */
      if (tcp->th_flags & TH_CWR) *p++ = 'C'; /* rfc 2481/3168 */
      *p++ = '\0';

      /* TCP Options */
      if((u32) tcp->th_off * 4 > sizeof(struct tcp_hdr)) {
        if(len < (u32) iphdrlen + (u32) tcp->th_off * 4 - frag_off) {
          Snprintf(tcpoptinfo, sizeof(tcpoptinfo), "option incomplete");
        } else {
          tcppacketoptinfo((u8*) tcp + sizeof(struct tcp_hdr),
                     tcp->th_off*4 - sizeof(struct tcp_hdr),
                     tcpoptinfo, sizeof(tcpoptinfo));
        }
      }

      /* Create a string with TCP information relevant to the specified level of detail */
      if( detail == LOW_DETAIL ){
      Snprintf(protoinfo, sizeof(protoinfo), "TCP %s:?? > %s:?? %s %s %s %s",
               srchost, dsthost, tflags, ipinfo, tcpinfo, tcpoptinfo);          
      }else if( detail == MEDIUM_DETAIL ){
      Snprintf(protoinfo, sizeof(protoinfo), "TCP %s:?? > %s:?? %s ack=%ul win=%hu %s IP [%s]",
               srchost, dsthost, tflags,
               ntohl(tcp->th_ack), ntohs(tcp->th_win),
               tcpoptinfo, ipinfo);
      }else if( detail==HIGH_DETAIL ){
          if( len >= (u32) iphdrlen + 12 ){ /* We have at least bytes 8-20 */
            Snprintf(protoinfo, sizeof(protoinfo), "TCP [%s:?? > %s:?? %s seq=%lu ack=%lu off=%d res=%d win=%hu csum=0x%04X urp=%d%s%s] IP [%s]",
               srchost, dsthost, tflags,
               (unsigned long) ntohl(tcp->th_seq),
               (unsigned long) ntohl(tcp->th_ack),
               (u8)tcp->th_off, (u8)tcp->th_x2, ntohs(tcp->th_win),
               ntohs(tcp->th_sum), ntohs(tcp->th_urp),
               (tcpoptinfo[0]!='\0') ? " " : "",
               tcpoptinfo, ipinfo );
           }else{ /* We only have bytes 8-16 */
                Snprintf(protoinfo, sizeof(protoinfo), "TCP %s:?? > %s:?? %s ack=%ul win=%hu %s IP [%s]",
                    srchost, dsthost, tflags,
                    ntohl(tcp->th_ack), ntohs(tcp->th_win),
                    tcpoptinfo, ipinfo);
            }
      }

    }
    
    /* CASE 3: where the IP packet is not a fragment but for some reason, we
     * don't have the entire TCP header, just part of it.*/
    else if ((len > (u32)iphdrlen) && (len < (u32) iphdrlen + 20)) {
     
        /* We only have the first 32 bits: source and dst port */
        if( (len >= (u32)iphdrlen + 4) && (len < (u32)iphdrlen + 8) ){
            Snprintf(protoinfo, sizeof(protoinfo), "TCP %s:%d > %s:%d ?? (incomplete) %s",
                     srchost, ntohs(tcp->th_sport), dsthost, ntohs(tcp->th_dport), ipinfo);
        }

         /* We only have the first 64 bits: ports and seq number */
        if( (len >= (u32)iphdrlen + 8) && (len < (u32)iphdrlen + 12) ) {
            Snprintf(tcpinfo, sizeof(tcpinfo), "TCP %s:%d > %s:%d ?? seq=%lu (incomplete) %s",
                     srchost, ntohs(tcp->th_sport), dsthost,
                     ntohs(tcp->th_dport), (unsigned long) ntohl(tcp->th_seq), ipinfo);
        }
        
        /* We only have the first 96 bits: ports, seq and ack number */
        if( (len >= (u32)iphdrlen + 12) && (len < (u32)iphdrlen + 16) ) {
            if(detail == LOW_DETAIL){ /* We don't print ACK in low detail */
                Snprintf(tcpinfo, sizeof(tcpinfo), "TCP %s:%d > %s:%d seq=%lu (incomplete), %s",
                         srchost, ntohs(tcp->th_sport), dsthost,
                         ntohs(tcp->th_dport), (unsigned long) ntohl(tcp->th_seq), ipinfo);
            }else{
                Snprintf(tcpinfo, sizeof(tcpinfo), "TCP [%s:%d > %s:%d seq=%lu ack=%lu (incomplete)] IP [%s]",
                         srchost, ntohs(tcp->th_sport), dsthost,
                         ntohs(tcp->th_dport), (unsigned long) ntohl(tcp->th_seq),
                         (unsigned long) ntohl(tcp->th_ack), ipinfo);
            }
        }

        /* We are missing the last 32 bits (checksum and urgent pointer) */
        if( (len >= (u32)iphdrlen + 16) && (len < (u32)iphdrlen + 20) ) {
            p = tflags;
            /* These are basically in tcpdump order */
            if (tcp->th_flags & TH_SYN) *p++ = 'S';
            if (tcp->th_flags & TH_FIN) *p++ = 'F';
            if (tcp->th_flags & TH_RST) *p++ = 'R';
            if (tcp->th_flags & TH_PUSH) *p++ = 'P';
            if (tcp->th_flags & TH_ACK){ *p++ = 'A';
            Snprintf(buf, sizeof(buf), " ack=%lu",
                 (unsigned long) ntohl(tcp->th_ack));
            strncat(tcpinfo, buf, sizeof(tcpinfo) - strlen(tcpinfo) - 1);
            }
            if (tcp->th_flags & TH_URG) *p++ = 'U';
            if (tcp->th_flags & TH_ECE) *p++ = 'E'; /* rfc 2481/3168 */
            if (tcp->th_flags & TH_CWR) *p++ = 'C'; /* rfc 2481/3168 */
            *p++ = '\0';


            /* Create a string with TCP information relevant to the specified level of detail */
            if(detail == LOW_DETAIL){ /* We don't print ACK in low detail */
                Snprintf(protoinfo, sizeof(protoinfo), "TCP %s:%d > %s:%d %s %s seq=%lu win=%hu (incomplete)",
                   srchost, ntohs(tcp->th_sport), dsthost, ntohs(tcp->th_dport),
                       tflags, ipinfo, (unsigned long) ntohl(tcp->th_seq),
                       ntohs(tcp->th_win));    
            }else if( detail == MEDIUM_DETAIL ){
                Snprintf(protoinfo, sizeof(protoinfo), "TCP [%s:%d > %s:%d %s seq=%lu ack=%lu win=%hu (incomplete)] IP [%s]",
                   srchost, ntohs(tcp->th_sport), dsthost, ntohs(tcp->th_dport),
                       tflags,  (unsigned long) ntohl(tcp->th_seq),
                       (unsigned long) ntohl(tcp->th_ack),
                       ntohs(tcp->th_win), ipinfo);    
            }else if( detail == HIGH_DETAIL ){
                Snprintf(protoinfo, sizeof(protoinfo), "TCP [%s:%d > %s:%d %s seq=%lu ack=%lu off=%d res=%d win=%hu (incomplete)] IP [%s]",
                   srchost, ntohs(tcp->th_sport),
                   dsthost, ntohs(tcp->th_dport),
                   tflags, (unsigned long) ntohl(tcp->th_seq),
                   (unsigned long) ntohl(tcp->th_ack),
                   (u8)tcp->th_off, (u8)tcp->th_x2, ntohs(tcp->th_win),
                   ipinfo );
            }

       }
    }
    
    /* CASE 4: where we (finally!) have a full 20 byte TCP header so we can
     * safely print all fields */
    else if (len >= (u32) iphdrlen + 20){
         
      /* TCP Flags */
      p = tflags;
      /* These are basically in tcpdump order */
      if (tcp->th_flags & TH_SYN) *p++ = 'S';
      if (tcp->th_flags & TH_FIN) *p++ = 'F';
      if (tcp->th_flags & TH_RST) *p++ = 'R';
      if (tcp->th_flags & TH_PUSH) *p++ = 'P';
      if (tcp->th_flags & TH_ACK){ *p++ = 'A';
        Snprintf(buf, sizeof(buf), " ack=%lu",
             (unsigned long) ntohl(tcp->th_ack));
        strncat(tcpinfo, buf, sizeof(tcpinfo) - strlen(tcpinfo) - 1);
      }
      if (tcp->th_flags & TH_URG) *p++ = 'U';
      if (tcp->th_flags & TH_ECE) *p++ = 'E'; /* rfc 2481/3168 */
      if (tcp->th_flags & TH_CWR) *p++ = 'C'; /* rfc 2481/3168 */
      *p++ = '\0';


      /* TCP Options */
      if((u32) tcp->th_off * 4 > sizeof(struct tcp_hdr)) {
        if(len < (u32) iphdrlen + (u32) tcp->th_off * 4) {
          Snprintf(tcpoptinfo, sizeof(tcpoptinfo), "option incomplete");

        } else {
          tcppacketoptinfo((u8*) tcp + sizeof(struct tcp_hdr),
                     tcp->th_off*4 - sizeof(struct tcp_hdr),
                     tcpoptinfo, sizeof(tcpoptinfo));
        }
      }

      /* Rest of header fields */
      if( detail == LOW_DETAIL ){
        Snprintf(protoinfo, sizeof(protoinfo), "TCP %s:%d > %s:%d %s %s seq=%lu win=%hu %s",
           srchost, ntohs(tcp->th_sport), dsthost, ntohs(tcp->th_dport),
               tflags, ipinfo, (unsigned long) ntohl(tcp->th_seq),
               ntohs(tcp->th_win), tcpoptinfo);               
      }else if( detail == MEDIUM_DETAIL ){
        Snprintf(protoinfo, sizeof(protoinfo), "TCP [%s:%d > %s:%d %s seq=%lu win=%hu csum=0x%04X%s%s] IP [%s]",
           srchost, ntohs(tcp->th_sport), dsthost, ntohs(tcp->th_dport),
               tflags, (unsigned long) ntohl(tcp->th_seq),
               ntohs(tcp->th_win),  ntohs(tcp->th_sum),
               (tcpoptinfo[0]!='\0') ? " " : "",
               tcpoptinfo, ipinfo );                 
      }else if( detail==HIGH_DETAIL ){
        Snprintf(protoinfo, sizeof(protoinfo), "TCP [%s:%d > %s:%d %s seq=%lu ack=%lu off=%d res=%d win=%hu csum=0x%04X urp=%d%s%s] IP [%s]",
           srchost, ntohs(tcp->th_sport),
           dsthost, ntohs(tcp->th_dport),
           tflags, (unsigned long) ntohl(tcp->th_seq),
           (unsigned long) ntohl(tcp->th_ack),
           (u8)tcp->th_off, (u8)tcp->th_x2, ntohs(tcp->th_win),
           ntohs(tcp->th_sum), ntohs(tcp->th_urp),
           (tcpoptinfo[0]!='\0') ? " " : "",
           tcpoptinfo, ipinfo );  
      }
    }
    else{
    /* If the packet does not fall into any other category, then we have a
     * really screwed up packet. */ 
        Snprintf(protoinfo, sizeof(protoinfo), "TCP %s:?? > %s:?? ?? %s (invalid TCP)",
        srchost, dsthost, ipinfo);
    }

    
/* UDP INFORMATION ***********************************************************/
  } else if (nextproto == IPPROTO_UDP && frag_off) {
      Snprintf(protoinfo, sizeof(protoinfo), "UDP %s:?? > %s:?? fragment %s (incomplete)", srchost, dsthost, ipinfo);
  } else if (nextproto == IPPROTO_UDP) {
    udp =  (struct udp_hdr *) (packet + sizeof(struct ip));
  /* TODO: See if we can segfault if we receive a fragmented packet whose IP packet does not say a thing about fragmentation */


  if( detail == LOW_DETAIL ){
    Snprintf(protoinfo, sizeof(protoinfo), "UDP %s:%d > %s:%d %s",
         srchost, ntohs(udp->uh_sport), dsthost, ntohs(udp->uh_dport),
         ipinfo);
  }else if( detail == MEDIUM_DETAIL ){
    Snprintf(protoinfo, sizeof(protoinfo), "UDP [%s:%d > %s:%d csum=0x%04X] IP [%s]",
         srchost, ntohs(udp->uh_sport), dsthost, ntohs(udp->uh_dport), ntohs(udp->uh_sum),
         ipinfo);
  }else if( detail==HIGH_DETAIL ){
    Snprintf(protoinfo, sizeof(protoinfo), "UDP [%s:%d > %s:%d len=%d csum=0x%04X] IP [%s]",
         srchost, ntohs(udp->uh_sport), dsthost, ntohs(udp->uh_dport),
         ntohs(udp->uh_ulen), ntohs(udp->uh_sum),
         ipinfo);
  }


/* SCTP INFORMATION **********************************************************/
  } else if (nextproto == IPPROTO_SCTP && frag_off) {
      Snprintf(protoinfo, sizeof(protoinfo), "SCTP %s:?? > %s:?? fragment %s (incomplete)", srchost, dsthost, ipinfo);
  } else if (nextproto == IPPROTO_SCTP) {
    sctp =  (struct sctp_hdr *) (packet + sizeof(struct ip));

    if( detail == LOW_DETAIL ){
        Snprintf(protoinfo, sizeof(protoinfo), "SCTP %s:%d > %s:%d %s",
             srchost, ntohs(sctp->sh_sport), dsthost, ntohs(sctp->sh_dport),
             ipinfo);
    }else if( detail == MEDIUM_DETAIL ){
        Snprintf(protoinfo, sizeof(protoinfo), "SCTP [%s:%d > %s:%d csum=0x%04x] IP [%s]",
             srchost, ntohs(sctp->sh_sport), dsthost, ntohs(sctp->sh_dport), ntohl(sctp->sh_sum),
             ipinfo);
    }else if( detail==HIGH_DETAIL ){
        Snprintf(protoinfo, sizeof(protoinfo), "SCTP [%s:%d > %s:%d vtag=%ul csum=0x%08x] IP [%s]",
             srchost, ntohs(sctp->sh_sport), dsthost, ntohs(sctp->sh_dport),
             ntohl(sctp->sh_sum), ntohl(sctp->sh_vtag),
             ipinfo);
    }


/* ICMP INFORMATION **********************************************************/
  } else if (nextproto == IPPROTO_ICMP && frag_off) {
      Snprintf(protoinfo, sizeof(protoinfo), "ICMP %s > %s fragment %s (incomplete)", srchost, dsthost, ipinfo);
  } else if (nextproto == IPPROTO_ICMP) {

    struct ip *ip2;       /* Points to the IP datagram carried by some ICMP messages */
    char *ip2dst;         /* Dest IP in caried IP datagram                   */
    u16 *nextmtu=NULL;    /* Store next hop MTU when ICMP==Frag required     */
    char auxbuff[128];    /* Aux buffer                                      */
    struct icmp_packet{   /* Generic ICMP struct */
      u8 type;
      u8 code;
      u16 checksum;
      u8 data[128];
    }*icmppkt;    
    struct ppkt {         /* Beginning of ICMP Echo/Timestamp header         */
      u8 type;
      u8 code;
      u16 checksum;
      u16 id;
      u16 seq;
    } *ping = NULL;
    struct icmp_redir{
      u8 type;
      u8 code;
      u16 checksum;
      u32 addr;
    } *icmpredir=NULL;
    struct icmp_router{
      u8 type;
      u8 code;
      u16 checksum;
      u8 addrs;
      u8 addrlen;
      u16 lifetime;
    } *icmprouter=NULL;
    struct icmp_param{
      u8 type;
      u8 code;
      u16 checksum;
      u8 pnt;
      u8 unused;
      u16 unused2;
    } *icmpparam=NULL;
    struct icmp_tstamp{
      u8 type;
      u8 code;
      u16 checksum;
      u16 id;
      u16 seq;
      u32 orig;
      u32 recv;
      u32 trans;
    } *icmptstamp=NULL;
    struct icmp_mask{
      u8 type;
      u8 code;
      u16 checksum;
      u16 id;
      u16 seq;
      u32 mask;
    } *icmpmask=NULL;
    
    /* Compute the length of the IP datagram + ICMP minimum length  */
    unsigned pktlen = (iphdrlen) + 8;

    /* We need the ICMP packet to be at least 8 bytes long */
    if (pktlen > len)
      goto icmpbad;
      
    ping = (struct ppkt *) ((iphdrlen) + (char *) ip);
    icmppkt=(struct icmp_packet *)  ((iphdrlen) + (char *) ip);
    
    switch(icmppkt->type) {

        /* Echo Reply **************************/
        case 0: 
          strcpy(icmptype, "Echo reply");
          Snprintf(icmpfields, sizeof(icmpfields), "id=%u seq=%u", ntohs(ping->id), ntohs(ping->seq) );
        break;

        /* Destination Unreachable *************/
        case 3: 
            /* Point to the start of the original datagram */
            ip2 = (struct ip *) ((char *) ip + (iphdrlen) + 8);
                
            /* Check we have a full IP datagram included in the ICMP message */
            pktlen += max( (ip2->ip_hl * 4), 20 );            
            if (pktlen > len){
                if(len==(u32)(iphdrlen) + 8 )
                    Snprintf(icmptype, sizeof icmptype, "Destination unreachable%s",
                    (detail!=LOW_DETAIL)? " (original datagram missing)" : "" );
                else
                    Snprintf(icmptype, sizeof icmptype, "Destination unreachable%s",
                    (detail!=LOW_DETAIL)? " (part of original datagram missing)" : "" );
                goto icmpbad;
            }

            /* Basic check to ensure we have an IPv4 datagram attached */
            /* TODO: We should actually check the datagram checksum to
             * see if it validates becuase just checking the version number
             * is not enough. On average, if we get random data 1 out of
             * 16 (2^4bits) times we will have value 4. */
            if( (ip2->ip_v != (u8)4) || ((ip2->ip_hl * 4)<20) || ((ip2->ip_hl * 4)>60) ){
                Snprintf(icmptype, sizeof icmptype, "Destination unreachable (bogus original datagram)");                         
                goto icmpbad;
            }else
            
            /* We have the original datagram + the first 8 bytes of the
             * transport layer header */    
            if ( (pktlen+8) < len) {
                tcp = (struct tcp_hdr *) ((char *) ip2 + (ip2->ip_hl * 4));
                udp = (struct udp_hdr *) ((char *) ip2 + (ip2->ip_hl * 4));
                sctp = (struct sctp_hdr *) ((char *) ip2 + (ip2->ip_hl * 4));
            }
            
            /* Determine the IP the original datagram was sent to */
            ip2dst = inet_ntoa(ip2->ip_dst);

            /* Determine type of Destination unreachable from the code value */
            switch (icmppkt->code) {
                case 0:
                    Snprintf(icmptype, sizeof icmptype, "Network %s unreachable", ip2dst);
                break;

                case 1:
                    Snprintf(icmptype, sizeof icmptype, "Host %s unreachable", ip2dst);
                break;

                case 2:
                    Snprintf(icmptype, sizeof icmptype, "Protocol %u unreachable", ip2->ip_p);
                break;

                case 3:
                    if ( (pktlen+8) < len){
                        if (ip2->ip_p == IPPROTO_UDP && udp)
                          Snprintf(icmptype, sizeof icmptype, "Port %u unreachable", ntohs(udp->uh_dport));
                        else if (ip2->ip_p == IPPROTO_TCP && tcp)
                          Snprintf(icmptype, sizeof icmptype, "Port %u unreachable", ntohs(tcp->th_dport));
                        else if (ip2->ip_p == IPPROTO_SCTP && sctp)
                          Snprintf(icmptype, sizeof icmptype, "Port %u unreachable", ntohs(sctp->sh_dport));
                        else
                          Snprintf(icmptype, sizeof icmptype, "Port unreachable (unknown protocol %u)", ip2->ip_p);
                    }
                    else
                        strcpy(icmptype, "Port unreachable");
                break;

                case 4:
                    strcpy(icmptype, "Fragmentation required");
                    nextmtu = (u16 *)(&(icmppkt->data[6]));
                    Snprintf(icmpfields, sizeof(icmpfields), "Next-Hop-MTU=%u", ntohs(*nextmtu) );
                break;
                
                case 5:
                    strcpy(icmptype, "Source route failed");
                break;
                
                case 6:
                    Snprintf(icmptype, sizeof icmptype, "Destination network %s unknown", ip2dst);
                break;

                case 7:
                    Snprintf(icmptype, sizeof icmptype, "Destination host %s unknown", ip2dst);
                break;
                
                case 8:
                    strcpy(icmptype, "Source host isolated");
                break;
                
                case 9:
                    Snprintf(icmptype, sizeof icmptype, "Destination network %s administratively prohibited", ip2dst);
                break;
                
                case 10:
                    Snprintf(icmptype, sizeof icmptype, "Destination host %s administratively prohibited", ip2dst);
                break;
                
                case 11:
                    Snprintf(icmptype, sizeof icmptype, "Network %s unreachable for TOS", ip2dst);
                break;
                
                case 12:
                    Snprintf(icmptype, sizeof icmptype, "Host %s unreachable for TOS", ip2dst);
                break;
                
                case 13:
                    strcpy(icmptype, "Communication administratively prohibited by filtering");
                break;
                
                case 14:
                    strcpy(icmptype, "Host precedence violation");
                break;
                
                case 15:
                    strcpy(icmptype, "Precedence cutoff in effect");
                break;
                
                default:
                    strcpy(icmptype, "Destination unreachable (unknown code)");
                break;
            } /* End of ICMP Code switch */
        break;


        /* Source Quench ***********************/        
        case 4:
            strcpy(icmptype, "Source quench");
        break;

        /* Redirect ****************************/
        case 5:
          if (ping->code == 0)
            strcpy(icmptype, "Network redirect");
          else if (ping->code == 1)
            strcpy(icmptype, "Host redirect");
          else strcpy(icmptype, "Redirect (unknown code)");
          icmpredir=(struct icmp_redir *)icmppkt;
          inet_ntop(AF_INET, &icmpredir->addr, auxbuff, sizeof(auxbuff) );
          Snprintf(icmpfields, sizeof(icmpfields), "addr=%s", auxbuff);                
        break;

        /* Echo Request ************************/        
        case 8:
          strcpy(icmptype, "Echo request");
          Snprintf(icmpfields, sizeof(icmpfields), "id=%u seq=%u", ntohs(ping->id), ntohs(ping->seq) );
        break;
        
        /* Router Advertisement ****************/
        case 9:
            if(icmppkt->code==16)
                strcpy(icmptype, "Router advertisement (Mobile Agent Only)");
            else
                strcpy(icmptype, "Router advertisement");
            icmprouter=(struct icmp_router *)icmppkt;          
            Snprintf(icmpfields, sizeof(icmpfields), "addrs=%u addrlen=%u lifetime=%d",
                icmprouter->addrs,
                icmprouter->addrlen,
                ntohs(icmprouter->lifetime) );
        break;

        /* Router Solicitation *****************/
        case 10:
          strcpy(icmptype, "Router solicitation");
        break;

        /* Time Exceeded ***********************/        
        case 11:
          if (icmppkt->code == 0)
            strcpy(icmptype, "TTL=0 during transit");
          else if (icmppkt->code == 1)
            strcpy(icmptype, "TTL=0 during reassembly");
          else strcpy(icmptype, "TTL exceeded (unknown code)");
        break;

        /* Parameter Problem *******************/        
        case 12:
          if (ping->code == 0)
            strcpy(icmptype, "Parameter problem (pointer indicates error)");
          else if (ping->code == 1)
            strcpy(icmptype, "Parameter problem (option missing)");
          else if (ping->code == 2)
            strcpy(icmptype, "Parameter problem (bad length)");
          else
            strcpy(icmptype, "Parameter problem (unknown code)");
          icmpparam=(struct icmp_param *)icmppkt;
          Snprintf(icmpfields, sizeof(icmpfields), "pointer=%d", icmpparam->pnt);            
        break;

        /* Timestamp Request/Reply *************/
        case 13:
        case 14:
          Snprintf(icmptype, sizeof(icmptype), "Timestamp %s", (icmppkt->type==13)? "request" : "reply" );
          icmptstamp=(struct icmp_tstamp *)icmppkt;
          Snprintf(icmpfields, sizeof(icmpfields), "id=%u seq=%u orig=%lu recv=%lu trans=%lu",
            ntohs(icmptstamp->id), ntohs(icmptstamp->seq),
            (unsigned long)ntohl(icmptstamp->orig),
            (unsigned long)ntohl(icmptstamp->recv),
            (unsigned long)ntohl(icmptstamp->trans) );     
        break;

        /* Information Request *****************/
        case 15:
          strcpy(icmptype, "Information request");
          Snprintf(icmpfields, sizeof(icmpfields), "id=%u seq=%u", ntohs(ping->id), ntohs(ping->seq) );
        break;

        /* Information Reply *******************/        
        case 16:
          strcpy(icmptype, "Information reply");
          Snprintf(icmpfields, sizeof(icmpfields), "id=%u seq=%u", ntohs(ping->id), ntohs(ping->seq) );
        break;

        /* Netmask Request/Reply ***************/ 
        case 17:
        case 18:
          Snprintf(icmptype, sizeof(icmptype), "Address mask %s", (icmppkt->type==17)? "request" : "reply" );
          icmpmask=(struct icmp_mask *)icmppkt;
          inet_ntop(AF_INET, &icmpmask->mask, auxbuff, sizeof(auxbuff) );
          Snprintf(icmpfields, sizeof(icmpfields), "id=%u seq=%u mask=%s",
            ntohs(ping->id), ntohs(ping->seq), auxbuff); 
        break;

        /* Traceroute **************************/ 
        case 30:
          strcpy(icmptype, "Traceroute");
        break;

        /* Domain Name Request *****************/ 
        case 37:
          strcpy(icmptype, "Domain name request");
        break;

        /* Domain Name Reply *******************/ 
        case 38:
          strcpy(icmptype, "Domain name reply");
        break;

        /* Security ****************************/ 
        case 40:
          strcpy(icmptype, "Security failures"); /* rfc 2521 */
        break;

        default:
          strcpy(icmptype, "Unknown type"); break;
        break;
    } /* End of ICMP Type switch */

    
    if (pktlen > len) {
    icmpbad:
      if (ping) {
        /* We still have this information */
        Snprintf(protoinfo, sizeof(protoinfo), "ICMP %s > %s %s (type=%d/code=%d) %s",
         srchost, dsthost, icmptype, ping->type, ping->code, ipinfo);
      } else {
        Snprintf(protoinfo, sizeof(protoinfo), "ICMP %s > %s [??] %s",
         srchost, dsthost, ipinfo);
      }
    } else {
            if(ping)
                sprintf(icmpinfo,"type=%d/code=%d", ping->type, ping->code);
            else
                strncpy(icmpinfo,"type=?/code=?", sizeof(icmpinfo) );        

        if( detail==LOW_DETAIL ){           
            Snprintf(protoinfo, sizeof(protoinfo), "ICMP %s > %s %s (%s) %s",
                srchost, dsthost, icmptype, icmpinfo, ipinfo);
        }else{
            Snprintf(protoinfo, sizeof(protoinfo), "ICMP [%s > %s %s (%s) %s] IP [%s]",
                srchost, dsthost, icmptype, icmpinfo, icmpfields, ipinfo);
        }
     
    }

/* UNKNOWN PROTOCOL **********************************************************/
  } else if( nextproto == IPPROTO_ICMPV6){
        Snprintf(protoinfo, sizeof(protoinfo), "ICMPv6 (%d) %s > %s: %s",
            ip->ip_p, srchost, dsthost, ipinfo);
  }else{

    if( nexthdrtoa(nextproto, true) == NULL ){
        Snprintf(protoinfo, sizeof(protoinfo), "Unknown protocol (%d) %s > %s: %s",
            ip->ip_p, srchost, dsthost, ipinfo);
    }else{
           Snprintf(protoinfo, sizeof(protoinfo), "%s (%d) %s > %s: %s",
            nexthdrtoa(nextproto, true), ip->ip_p, srchost, dsthost, ipinfo); 
    }
  }

  return protoinfo;
}
/*****************************************************************************/

/*
  if( detail == LOW_DETAIL ){
  }else if( detail == MEDIUM_DETAIL ){
  }else if( detail==HIGH_DETAIL ){
  }
*/





/*****************************************************************************/
char *ll2shortascii(unsigned long long bytes, char *buf, int buflen) {
  if (buflen < 2 || !buf) fatal("Bogus parameter passed to %s", __func__);

  if (bytes > 1000000) {
    Snprintf(buf, buflen, "%.3fMB", bytes / 1000000.0);
  } else if (bytes > 10000) {
    Snprintf(buf, buflen, "%.3fKB", bytes / 1000.0);
  } else Snprintf(buf, buflen, "%uB", (unsigned int) bytes);

  return buf;
}
/*****************************************************************************/




/*****************************************************************************/
/* CHANGES: Commented out PacketTrace::... line */
/* Issues an ARP request for the MAC of targetss (which will be placed
   in targetmac if obtained) from the source IP (srcip) and source mac
   (srcmac) given.  "The request is ussued using device dev to the
   broadcast MAC address.  The transmission is attempted up to 3
   times.  If none of these elicit a response, false will be returned.
   If the mac is determined, true is returned. */
bool doArp(const char *dev, const u8 *srcmac, 
	   const struct sockaddr_storage *srcip, 
	   const struct sockaddr_storage *targetip, u8 *targetmac) {
  /* timeouts in microseconds ... the first ones are retransmit times, while 
     the final one is when we give up */
  int timeouts[] = { 100000, 400000, 800000 }; 
  int max_sends = 3;
  int num_sends = 0; // How many we have sent so far 
  eth_t *ethsd;
  u8 frame[ETH_HDR_LEN + ARP_HDR_LEN + ARP_ETHIP_LEN];
  const struct sockaddr_in *targetsin = (struct sockaddr_in *) targetip;
  const struct sockaddr_in *srcsin = (struct sockaddr_in *) srcip;
  struct timeval start, now, rcvdtime;
  int timeleft;
  int listenrounds;
  int rc;
  pcap_t *pd;
  struct in_addr rcvdIP;
  bool foundit = false;

  if (targetsin->sin_family != AF_INET || srcsin->sin_family != AF_INET)
    fatal("%s can only handle IPv4 addresses", __func__);

  /* Start listening */
  pd = my_pcap_open_live(dev, 50, 1, 25);
  set_pcap_filter(dev, pd, "arp and ether dst host %02X:%02X:%02X:%02X:%02X:%02X", srcmac[0], srcmac[1], srcmac[2], srcmac[3], srcmac[4], srcmac[5]);

  /* Prepare probe and sending stuff */
  ethsd = eth_open_cached(dev);
  if (!ethsd) fatal("%s: failed to open device %s", __func__, dev);
  eth_pack_hdr(frame, ETH_ADDR_BROADCAST, *srcmac, ETH_TYPE_ARP);
  arp_pack_hdr_ethip(frame + ETH_HDR_LEN, ARP_OP_REQUEST, *srcmac, 
		     srcsin->sin_addr, ETH_ADDR_BROADCAST, 
		     targetsin->sin_addr);
  gettimeofday(&start, NULL);
  gettimeofday(&now, NULL);

  while(!foundit && num_sends < max_sends) {
    /* Send the sucker */
    rc = eth_send(ethsd, frame, sizeof(frame));
    if (rc != sizeof(frame)) {
      error("WARNING: %s: eth_send of ARP packet returned %u rather than expected %d bytes", __func__, rc, (int) sizeof(frame));
    }
    //PacketTrace::traceArp(PacketTrace::SENT, (u8 *) frame, sizeof(frame), &now);
    num_sends++;
    
    listenrounds = 0;
    while(!foundit) {
      gettimeofday(&now, NULL);
      timeleft = timeouts[num_sends - 1] - TIMEVAL_SUBTRACT(now, start);
      if (timeleft < 0) {
	if (listenrounds > 0) break;
	else timeleft = 25000;
      }
      listenrounds++;
      /* Now listen until we reach our next timeout or get an answer */
      rc = read_arp_reply_pcap(pd, targetmac, &rcvdIP, timeleft, &rcvdtime);
      if (rc == -1) fatal("%s: Received -1 response from readarp_reply_pcap", 
			  __func__);
      if (rc == 1) {
	/* Yay, I got one! But is it the right one? */
	if (rcvdIP.s_addr != targetsin->sin_addr.s_addr)
	  continue;  /* D'oh! */
	foundit = true;  /* WOOHOO! */
      }
    }
  }

  /* OK - let's close up shop ... */
  pcap_close(pd);
  /* No need to close ethsd due to caching */
  return foundit;
}
/*****************************************************************************/





/*****************************************************************************/
/* CHANGES: commented out reference to o.debugging and log_write() */
/* Set a pcap filter */
void set_pcap_filter(const char *device,
		     pcap_t *pd, const char *bpf, ...)
{
  va_list ap;
  char buf[3072];
  struct bpf_program fcode;
#ifndef __amigaos__
  unsigned int localnet, netmask;
#else
  bpf_u_int32 localnet, netmask;
#endif
  char err0r[256];
  
  // Cast below is becaue OpenBSD apparently has a version that takes a
  // non-const device (hopefully they don't actually write to it).
  if (pcap_lookupnet( (char *) device, &localnet, &netmask, err0r) < 0)
    fatal("Failed to lookup subnet/netmask for device (%s): %s", device, err0r);
  
  va_start(ap, bpf);
  if (Vsnprintf(buf, sizeof(buf), bpf, ap) >= (int) sizeof(buf))
    fatal("%s called with too-large filter arg\n", __func__);
  va_end(ap);

  /* Due to apparent bug in libpcap */
  /* Maybe this bug no longer exists ... I'll comment out for now 
   *      if (islocalhost(target->v4hostip()))
   *      buf[0] = '\0'; */

  //if (o.debugging)
  //  log_write(LOG_STDOUT, "Packet capture filter (device %s): %s\n", device, buf);
  
  if (pcap_compile(pd, &fcode, buf, 0, netmask) < 0)
    fatal("Error compiling our pcap filter: %s", pcap_geterr(pd));
  if (pcap_setfilter(pd, &fcode) < 0 )
    fatal("Failed to set the pcap filter: %s\n", pcap_geterr(pd));
  pcap_freecode(&fcode);
}
/*****************************************************************************/




/*****************************************************************************/
/* CHANGES: Commented out PacketTrace::... line */
/* Attempts to read one IPv4/Ethernet ARP reply packet from the pcap
   descriptor pd.  If it receives one, fills in sendermac (must pass
   in 6 bytes), senderIP, and rcvdtime (can be NULL if you don't care)
   and returns 1.  If it times out and reads no arp requests, returns
   0.  to_usec is the timeout periaod in microseconds.  Use 0 to avoid
   blocking to the extent possible.  Returns
   -1 or exits if ther is an error. */
int read_arp_reply_pcap(pcap_t *pd, u8 *sendermac, struct in_addr *senderIP,
		       long to_usec, struct timeval *rcvdtime) {
  static int warning = 0;
  int datalink;
  struct pcap_pkthdr head;
  u8 *p;
  int timedout = 0;
  int badcounter = 0;
  struct timeval tv_start, tv_end;

  if (!pd) fatal("NULL packet device passed to %s", __func__);

  if (to_usec < 0) {
    if (!warning) {
      warning = 1;
      error("WARNING: Negative timeout value (%lu) passed to %s() -- using 0", to_usec, __func__);
    }
    to_usec = 0;
  }

  /* New packet capture device, need to recompute offset */
  if ( (datalink = pcap_datalink(pd)) < 0)
    fatal("Cannot obtain datalink information: %s", pcap_geterr(pd));

  if (datalink != DLT_EN10MB)
    fatal("%s called on interfaces that is datatype %d rather than DLT_EN10MB (%d)", __func__, datalink, DLT_EN10MB);

  if (to_usec > 0) {
    gettimeofday(&tv_start, NULL);
  }

  do {
#ifdef WIN32
    if (to_usec == 0)
      PacketSetReadTimeout(pd->adapter, 1);
    else {
      gettimeofday(&tv_end, NULL);
      long to_left = MAX(1, (to_usec - TIMEVAL_SUBTRACT(tv_end, tv_start)) / 1000);
      // Set the timeout (BUGBUG: this is cheating)
      PacketSetReadTimeout(pd->adapter, to_left);
    }
#endif

   p = NULL;

   if (pcap_select(pd, to_usec) == 0)
     timedout = 1;
   else
     p = (u8 *) pcap_next(pd, &head);

    if (p && head.caplen >= 42) { /* >= because Ethernet padding makes 60 */
      /* frame type 0x0806 (arp), hw type eth (0x0001), prot ip (0x0800),
	 hw size (0x06), prot size (0x04) */
      if (memcmp(p + 12, "\x08\x06\x00\x01\x08\x00\x06\x04\x00\x02", 10) == 0) {
	memcpy(sendermac, p + 22, 6);
	/* I think alignment should allow this ... */
	memcpy(&senderIP->s_addr, p+28, 4);
	break;
      }
    }

    if (!p) {
      /* Should we timeout? */
      if (to_usec == 0) {
	timedout = 1;
      } else if (to_usec > 0) {
	gettimeofday(&tv_end, NULL);
	if (TIMEVAL_SUBTRACT(tv_end, tv_start) >= to_usec) {
	  timedout = 1;     
	}
      }
    } else {
      /* We'll be a bit patient if we're getting actual packets back, but
	 not indefinitely so */
      if (badcounter++ > 50)
	timedout = 1;
    }
  } while(!timedout);

  if (timedout) return 0;

  if (rcvdtime) {
    // FIXME: I eventually need to figure out why Windows head.ts time is sometimes BEFORE the time I
    // sent the packet (which is according to gettimeofday() in nbase).  For now, I will sadly have to
    // use gettimeofday() for Windows in this case
    // Actually I now allow .05 discrepancy.   So maybe this isn't needed.  I'll comment out for now.
    // Nope: it is still needed at least for Windows.  Sometimes the time from he pcap header is a 
    // COUPLE SECONDS before the gettimeofday() results :(.
#if defined(WIN32) || defined(__amigaos__)
    gettimeofday(&tv_end, NULL);
    *rcvdtime = tv_end;
#else
    rcvdtime->tv_sec = head.ts.tv_sec;
    rcvdtime->tv_usec = head.ts.tv_usec;
    assert(head.ts.tv_sec);
#endif
  }
  //PacketTrace::traceArp(PacketTrace::RCVD, (u8 *) p, 42,  rcvdtime);

  return 1;
}


/* IPv6 compatible version of nmap's devname2ipaddr()
 * @warning For this to work we need getinterfaces() not to skip IPv6 */
int devname2ipaddr(char *dev, struct sockaddr_storage *addr) {
struct interface_info *mydevs;
struct sockaddr_storage *s=NULL;
struct sockaddr_in *s4=NULL;
struct sockaddr_in6 *s6=NULL;
int numdevs;
int i;
mydevs = getinterfaces(&numdevs);

if (!mydevs) return -1;

if( !addr || !dev )
    fatal("devname2ipaddr(): NULL values supplied.");

  for(i=0; i < numdevs; i++) {
    s=(struct sockaddr_storage *)&mydevs[i].addr;
    s4=(struct sockaddr_in *)&mydevs[i].addr;
    s6=(struct sockaddr_in6 *)&mydevs[i].addr;
    if (s4->sin_family==AF_INET || s6->sin6_family==AF_INET6){
        if (!strcmp(dev, mydevs[i].devfullname)) {
            memcpy(addr, s, sizeof(struct sockaddr_storage));
            return 0;
        }    
    } else{ /* Unknown family, skipping it... */
      continue;
    }
  }
  return -1;
  
} /* End of devname2ipaddr() */


#ifndef WIN32
struct interface_info *getinterfaces_siocgifconf_ipv6(int *howmany) {
  struct interface_info *devs;
  int count = 0;
  int capacity = 0;
  struct ifconf ifc;
  struct ifreq *ifr;
  int sd;
  int len;

  capacity = 16;
  devs = (struct interface_info *) safe_zalloc(sizeof(struct interface_info) * capacity);

  /* Dummy socket for ioctl */
  sd = socket(AF_INET6, SOCK_DGRAM, 0);
  if (sd < 0)
    pfatal("socket in %s", __func__);

  ifc.ifc_len = 20480;
  ifc.ifc_buf = (char *) safe_zalloc(ifc.ifc_len);
  /* Returns an array of struct ifreq in ifc.ifc_req, which is a union with
     ifc.ifc_buf. */
  if (ioctl(sd, SIOCGIFCONF, &ifc) < 0)
    fatal("Failed to determine your configured interfaces!\n");
  if (ifc.ifc_len == 0) 
    fatal("%s: SIOCGIFCONF claims you have no network interfaces!\n", __func__);
  else
    printf("DEBUGGGGGGGGGGGGGGG:  %d interfaces\n", ifc.ifc_len);;
  ifr = ifc.ifc_req;

  for(ifr = ifc.ifc_req;
      ifr && ifr->ifr_name[0] && (void *) ifr < (char *) ifc.ifc_buf + ifc.ifc_len;
      ifr = (struct ifreq *) ((char *) ifr + len)) {
    struct sockaddr_in *sin4;
    struct sockaddr_in6 *sin6;
    struct ifreq tmpifr;
    u16 ifflags;
    int rc;
    char *p;

    /* On some platforms (such as FreeBSD), the length of each ifr changes
       based on the sockaddr type used, so we get the next length now. */
#if HAVE_SOCKADDR_SA_LEN
    len = ifr->ifr_addr.sa_len + sizeof(ifr->ifr_name);
#else
    len = sizeof(struct ifreq);
#endif

    /* skip any device with no name */
    if (ifr->ifr_name[0] == '\0')
      continue;


    sin4 = (struct sockaddr_in *) &ifr->ifr_addr;
    sin6 = (struct sockaddr_in6 *) &ifr->ifr_addr;
    /* Skip unknown address families */
    if (sin4->sin_family!=AF_INET && sin6->sin6_family!=AF_INET6)
      continue;

    /* Make room for this new interface if necessary. */
    if (count >= capacity) {
      capacity <<= 2;
      devs = (struct interface_info *) safe_realloc(devs, sizeof(struct interface_info) * capacity);
    }

    /* We know the address, put it in the array. */
    if (sin4->sin_family==AF_INET)
        memcpy(&(devs[count].addr), sin4, MIN(sizeof(devs[count].addr), sizeof(*sin4)));
    else
        memcpy(&(devs[count].addr), sin6, MIN(sizeof(devs[count].addr), sizeof(*sin6)));
        
    Strncpy(devs[count].devname, ifr->ifr_name, sizeof(devs[count].devname));
    Strncpy(devs[count].devfullname, ifr->ifr_name, sizeof(devs[count].devfullname));
    /* devname isn't allowed to have alias qualification */
    p = strchr(devs[count].devname, ':');
    if (p != NULL)
      *p = '\0';

    /* Use tmpifr for further ioctl requests. We're going to make a bunch of
       ioctl calls to learn about the interface and set fields in devs[count].

       The Linux netdevice(7) man page says that you only have to set ifr_name
       before making the ioctl, but perhaps other platforms need ifr_addr to be
       set too. ifr_name will persist but ifr_addr is in a union with the ioctl
       return value, so it has to be reset before every call. The general
       pattern is memcpy, then ioctl. */
    Strncpy(tmpifr.ifr_name, ifr->ifr_name, sizeof(tmpifr.ifr_name));

    /* Look up the netmask. Note setting of ifr_addr. */
    if (sin4->sin_family==AF_INET)
        memcpy(&tmpifr.ifr_addr, sin4, MIN(sizeof(tmpifr.ifr_addr), sizeof(*sin4)));
    else
        memcpy(&tmpifr.ifr_addr, sin6, MIN(sizeof(tmpifr.ifr_addr), sizeof(*sin6)));
        
    rc = ioctl(sd, SIOCGIFNETMASK, &tmpifr);
    if (rc < 0 && errno != EADDRNOTAVAIL)
            printf("Failed to determine the netmask of %s!\n", tmpifr.ifr_name);
    else if (rc < 0)
      devs[count].netmask_bits = 32;
    else {
      /* We would use ifr_netmask, but that's only on Linux, so use ifr_addr
         which shares the same memory space in a union. */
      addr_stob(&(tmpifr.ifr_addr), &devs[count].netmask_bits);
    }

    /* Now we need to determine the device type ... this technique is kinda iffy
       ... may not be portable. */
    /* Get the flags. */
    if (sin4->sin_family==AF_INET)
        memcpy(&tmpifr.ifr_addr, sin4, MIN(sizeof(tmpifr.ifr_addr), sizeof(*sin4)));
    else
        memcpy(&tmpifr.ifr_addr, sin6, MIN(sizeof(tmpifr.ifr_addr), sizeof(*sin6)));
    rc = ioctl(sd, SIOCGIFFLAGS, &tmpifr);
    if (rc < 0)
      pfatal("Failed to get IF Flags for device %s", ifr->ifr_name);

    ifflags = tmpifr.ifr_flags;

    if (ifflags & IFF_LOOPBACK) {
      devs[count].device_type = devt_loopback;
    } else if (ifflags & IFF_BROADCAST) {
      devs[count].device_type = devt_ethernet;

      /* If the device type is ethernet, get the MAC address. */
#ifdef SIOCGIFHWADDR
      if (sin4->sin_family==AF_INET)
          memcpy(&tmpifr.ifr_addr, sin4, MIN(sizeof(tmpifr.ifr_addr), sizeof(*sin4)));
      else
          memcpy(&tmpifr.ifr_addr, sin6, MIN(sizeof(tmpifr.ifr_addr), sizeof(*sin6)));
      rc = ioctl(sd, SIOCGIFHWADDR, &tmpifr);
      if (rc < 0 && errno != EADDRNOTAVAIL)
        pfatal("Failed to determine the MAC address of %s!", tmpifr.ifr_name);
      else if (rc >= 0)
        memcpy(devs[count].mac, &tmpifr.ifr_addr.sa_data, 6);
#else
      /* Let's just let libdnet handle it ... */
      eth_t *ethsd = eth_open_cached(devs[count].devname);
      eth_addr_t ethaddr;

      if (!ethsd) {
        error("Warning: Unable to open interface %s -- skipping it.", devs[count].devname);
        continue;
      }
      if (eth_get(ethsd, &ethaddr) != 0) {
        error("Warning: Unable to get hardware address for interface %s -- skipping it.", devs[count].devname);
        continue;
      }
      memcpy(devs[count].mac, ethaddr.data, 6);
#endif /*SIOCGIFHWADDR*/
    } else if (ifflags & IFF_POINTOPOINT) {
      devs[count].device_type = devt_p2p;
    } else {
      devs[count].device_type = devt_other;
    }

    if (ifflags & IFF_UP)
      devs[count].device_up = true;
    else
      devs[count].device_up = false;

    /* All done with this interface. Increase the count. */
    count++;
  }
  free(ifc.ifc_buf);
  close(sd);

  *howmany = count;
  return devs;
}
#endif

